# inbuilt
import os
import json

# external
import pandas as pd

# internal
from handuflow.validation.ValidationRule import ValidationRule
from handuflow.validation.ValidationContext import ValidationContext
from handuflow.exception.ValidationError import ValidationError

class EnforceMasterSpecsStructure(ValidationRule):
    name = "Enforce master spec structure"
    def validate(self, context: ValidationContext):
        master_specs_path = os.path.join(context.file_hunt_path, context.master_spec_name)
        if os.path.exists(master_specs_path) == False:
            raise ValidationError(
                message = f"System can't find the Master Specs File at [{context.file_hunt_path}]. Terminating process", 
                original_exception=None
            )
        context.master_specs_dataframe = pd.read_excel(master_specs_path, sheet_name="master_specs")
        context.master_specs_dataframe = context.master_specs_dataframe[context.master_specs_dataframe['is_active'] == True]
        cols = context.master_specs_dataframe.columns.tolist()
        required_columns = [
            "feed_id",
            "system_name",
            "subsystem_name",
            "category",
            "sub_category",
            "data_flow_direction",
            "residing_layer",
            "feed_name",
            "feed_type",
            "feed_specs",
            "load_type",
            "target_unity_catalog",
            "target_schema_name",
            "target_table_name",
            "suggested_feed_name",
            "parallelism_group_number",
            "parent_feed_id",
            "is_active",
        ]
        expected = set(required_columns)
        actual = set(context.master_specs_dataframe.columns)
        missing = expected - actual
        extra = actual - expected
        if missing or extra:
            raise ValidationError(
                message = f"Missing columns: {sorted(missing)} | Extra columns: {sorted(extra)}", 
                original_exception=None
            )
        subset = context.master_specs_dataframe[required_columns].replace(r"^\s*$", pd.NA, regex=True)
        null_mask = subset.isna()
        if null_mask.any().any():
            errors = (
                null_mask
                .stack()
                .reset_index()
            )
            errors.columns = ["row_index", "column", "is_null"]
            errors = errors[errors["is_null"]]
            errors["row_number"] = errors["row_index"] + 2
            error_lines = []
            grouped = (
                errors[["row_number", "column"]]
                .groupby("row_number")["column"]
                .apply(list)
            )
            for row_number, columns in grouped.items():
                cols = ", ".join(columns)
                error_lines.append(f"  • Row {row_number}: {cols}")
            message = (
                "Validation failed: required columns contain null or blank values\n\n"
                "Affected rows and columns:\n"
                + "\n".join(error_lines) + '\n'
            )
            raise ValidationError(
                message=message,
                original_exception=None            
            )
        else:

            filtered_df = context.master_specs_dataframe[
                context.master_specs_dataframe['data_flow_direction'] != 'SOURCE_TO_BRONZE'
            ]

            all_feed_specs = filtered_df['feed_specs'].to_list()
            for feed_specs in all_feed_specs:
                feed_specs_dict = json.loads(feed_specs)
                required_keys = [
                    "primary_key",
                    "composite_key",
                    "partition_keys",
                    "vacuum_hours",
                    "source_table_name",
                    "selection_query",
                    "selection_schema",
                    "standard_checks",
                    "comprehensive_checks"
                ]
                top_level_keys = list(feed_specs_dict.keys())
                if set(required_keys) != set(top_level_keys):
                    missing_keys = set(required_keys) - set(top_level_keys)
                    extra_keys = set(top_level_keys) - set(required_keys)
                    message = f"Validation failed! Missing keys: {missing_keys}"
                    if extra_keys:
                        message += f"; Unexpected keys: {extra_keys}"
                    raise ValidationError(
                        message=message,
                        original_exception=None
                    )











# inbuilt
import logging

# external
import pyspark.sql.functions as F
from pyspark.sql import SparkSession
from pyspark.sql.types import StructType
from delta.tables import DeltaTable

# internal
from handuflow.data_movement_controller.data_class.LoadResult import LoadResult
from handuflow.data_movement_controller.data_class.LoadConfig import LoadConfig
from handuflow.data_movement_controller.BaseLoadStrategy import BaseLoadStrategy
from handuflow.exception.DataLoadException import DataLoadException

class SCDType2(BaseLoadStrategy):

    def __init__(self, config: LoadConfig, spark: SparkSession) -> None:
        super().__init__(config=config, spark=spark)
        self.logger = logging.getLogger(__name__)
        self.logger.info("Initializing SCD_TYPE_2 data transfer component...")

    def __supports_identity(self) -> bool:
        return "databricks" in self.spark.version.lower()

    def load(self):
        try:
            self.logger.info("Perform SCD Type-2 load using a true Delta IDENTITY surrogate key.")
            self.logger.info("Executing SCD Type-2 load.")
            self._enforce_load_type_consistency()
            if not self._create_staging_layer():
                self.logger.warning("Incremental CDC aborted: staging layer not created.")
                return LoadResult(
                    feed_id = self.config.master_specs['feed_id'], 
                    success=False,
                    total_rows_inserted=0, 
                    total_rows_deleted=0, 
                    total_rows_updated=0
                )
            incr_df = self._current_staging_incremental_table_df
            incr_df = incr_df.drop("_x_row_hash", "_x_commit_version", "_x_commit_timestamp")
            if incr_df is None or incr_df.count() == 0:
                self.logger.warning(f"No CDC changes found for {self._current_target_table_name}.")
                return LoadResult(
                    feed_id = self.config.master_specs['feed_id'], 
                    success=False,
                    skipped=True,
                    total_rows_inserted=0, 
                    total_rows_deleted=0, 
                    total_rows_updated=0
                )
            target_table = self._current_target_table_name
            primary_key = self.config.feed_specs["primary_key"]
            composite_keys = self.config.feed_specs["composite_key"]
            all_keys = [primary_key] if primary_key else []
            all_keys.extend([k for k in composite_keys if k not in all_keys])
            if not all_keys:
                self.logger.error(f"No keys defined for SCD2 load: {target_table}")
                return LoadResult(
                    feed_id = self.config.master_specs['feed_id'], 
                    success=False,
                    total_rows_inserted=0, 
                    total_rows_deleted=0, 
                    total_rows_updated=0
                )
            non_key_columns = [
                c for c in incr_df.columns
                if c not in all_keys and c not in ["_x_operation"]
            ]
            sort_cols = [F.col(k).asc() for k in all_keys]
            incr_df = incr_df.orderBy(*sort_cols)
            incr_df = incr_df.withColumn(
                "_x_row_hash",
                F.sha2(
                    F.concat_ws("||", *[F.col(c).cast("string") for c in non_key_columns]),
                    256
                )
            )
            incr_df = incr_df.withColumn("_x_date_from", F.current_timestamp())
            incr_df = incr_df.withColumn("_x_date_to", F.to_timestamp(F.lit("9999-12-31 23:59:59")))
            incr_df = incr_df.withColumn("_x_is_active", F.lit(1))
            incr_df = incr_df.withColumn("_x_last_modification_timestamp", F.current_timestamp())
            incr_df = incr_df.withColumn("_x_last_operation", F.col("_x_operation"))
            if not self.spark.catalog.tableExists(target_table):
                schema_cols = [
                    f"{field.name} {field.dataType.simpleString()}"
                    for field in incr_df.schema
                    if field.name not in ["_x_operation"]  # safety
                ]
                if self.__supports_identity():
                    surrogate_col = "_x_surrogate_key BIGINT GENERATED BY DEFAULT AS IDENTITY"
                else:
                    surrogate_col = "_x_surrogate_key BIGINT"
                ddl = f"""
                CREATE TABLE {target_table} (
                    {surrogate_col},
                    {", ".join(schema_cols)}
                ) USING DELTA
                """
                self.spark.sql(ddl)
                incr_df = self._enforce_schema(incr_df, StructType.fromJson(self.config.feed_specs['selection_schema']))
                (
                    incr_df
                        .drop("_x_operation")
                        .write
                        .format("delta")
                        .mode("append")
                        .saveAsTable(target_table)
                )
                self.logger.info(f"Created SCD-2 table {target_table} with IDENTITY surrogate key.")
                return LoadResult(
                    feed_id = self.config.master_specs['feed_id'], 
                    success=True,
                    total_rows_inserted=0, 
                    total_rows_deleted=0, 
                    total_rows_updated=0
                )
            target = DeltaTable.forName(self.spark, target_table)
            key_cond = " AND ".join([f"t.{k} = s.{k}" for k in all_keys])
            close_cond = (
                f"{key_cond} AND t._x_is_active = 1 AND ("
                f"t._x_row_hash <> s._x_row_hash OR s._x_operation = 'delete'"
                ")"
            )
            (
                target.alias("t")
                    .merge(incr_df.alias("s"), close_cond)
                    .whenMatchedUpdate(set={
                            "_x_is_active": "0",
                            "_x_date_to": "current_timestamp()",
                            "_x_last_modification_timestamp": "current_timestamp()",
                            "_x_last_operation": "s._x_operation"
                    })
                    .execute()
            )
            insert_on = f"{key_cond} AND t._x_is_active = 1 AND t._x_row_hash = s._x_row_hash"
            incr_df = self._enforce_schema(incr_df, StructType.fromJson(self.config.feed_specs['selection_schema']))
            (
                target.alias("t")
                .merge(incr_df.alias("s"), insert_on)
                .whenNotMatchedInsert(
                    condition="s._x_operation != 'delete'",   # do not insert deletes
                    values={
                        c: f"s.{c}"
                        for c in incr_df.columns
                        if c not in ["_x_commit_version", "_x_commit_timestamp", "_x_operation"]  # drop ETL metadata
                    }
                )
                .execute()
            )
            self.spark.sql(
                f"ALTER TABLE {target_table} "
                f"SET TBLPROPERTIES ('data.load_type' = '{self.config.master_specs["load_type"]}')"
            )
            self.logger.info(f"SCD-2 load completed: {target_table}")
            return LoadResult(
                feed_id = self.config.master_specs['feed_id'], 
                success=True,
                total_rows_inserted=0, 
                total_rows_deleted=0, 
                total_rows_updated=0
            )
        except Exception as e:
            raise DataLoadException(
                original_exception=e,
                message=f"Error during SCD_TYPE_2 for {self._current_target_table_name}: {str(e)}"           
            )


# logging.py
import logging
import sys
from datetime import datetime, UTC

class LoggingPrettyFormatter(logging.Formatter):
    """Color-coded formatter for root logger."""

    COLORS = {
        "INFO": "\033[94m",      # blue
        "WARNING": "\033[93m",   # yellow
        "ERROR": "\033[91m",     # red
        "CRITICAL": "\033[95m",  # magenta
        "ENDC": "\033[0m",
    }

    
    def __init__(self, is_log_verbose: bool = True):
        super().__init__()
        self.is_log_verbose = is_log_verbose


    def format(self, record: logging.LogRecord) -> str:
        ts = datetime.now(UTC).strftime("%Y-%m-%d %H:%M:%S")

        # Disable colors if not a real terminal (Databricks, logs, files)
        if sys.stdout.isatty():
            color = self.COLORS.get(record.levelname, "")
            endc = self.COLORS["ENDC"]
        else:
            color = endc = ""

        message = record.getMessage()
        return f"{color}[{ts}] {f"[{record.name}]" if self.is_log_verbose else ""} [{record.levelname}] - {message}{endc}"


# inbuilt
import os
import uuid
import time
import random
import logging
import requests
from io import BytesIO
from requests.exceptions import RequestException

# external
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType

# internal
from handuflow.data_movement_controller.BaseLoadStrategy import BaseLoadStrategy
from handuflow.data_movement_controller.data_class.LoadConfig import LoadConfig
from handuflow.data_movement_controller.data_class.LoadResult import LoadResult
from handuflow.exception.ExtractionException import ExtractionException


class APIExtractor(BaseLoadStrategy):

    def __init__(self, config: LoadConfig, spark: SparkSession) -> None:
        super().__init__(config=config, spark=spark)
        self.logger = logging.getLogger(__name__)
        self.config = config
        self.spark = spark
        if self.config.target_unity_catalog == "testing":
            self.__bronze_schema = f"bronze"
        else:
            self.__bronze_schema = f"{self.config.target_unity_catalog}.bronze"
        
        self.logger.warning('API Extractor will always dump data in bronze schema as per medallion architecture.')

    def load(self) -> LoadResult:
        try:
            result_df = self.__extract()
            self.spark.sql(f"CREATE SCHEMA IF NOT EXISTS {self.__bronze_schema}")
            feed_temp = (
                f"{self.__bronze_schema}."
                f"{self.config.master_specs['target_table_name']}"
            )
            self.logger.info(f"Creating bronze table: {feed_temp}")
            schema = StructType.fromJson(self.config.feed_specs['selection_schema'])
            result_df = self._enforce_schema(result_df, schema)
            (
                result_df.write.
                format("delta")
                .mode("overwrite")
                .saveAsTable(feed_temp)
            )
            return LoadResult(
                feed_id = self.config.master_specs['feed_id'],
                success=True, 
                total_rows_inserted=result_df.count(),
                total_rows_updated=0,
                total_rows_deleted=0
            )
        except Exception as e:
            raise ExtractionException(
                message=f"Feed ID: {self.config.master_specs['feed_id']}, Error during FULL LOAD for {self._current_target_table_name}: {str(e)}",
                original_exception=e
            )
    
    def __validate_content_type(self, response: requests.Response, expected: str) -> None:
        content_type = response.headers.get("Content-Type", "").lower()
        if expected == "json" and "application/json" not in content_type:
            raise ValueError(f"Expected JSON but got {content_type}")
        if expected == "parquet":
            if not content_type.startswith("application/parquet"):
                raise ValueError(f"Expected Parquet but got {content_type}")

    def __extract(self) -> DataFrame:
        try:
            cfg = self.config.feed_specs["ingestion_config"]
            response_type = cfg['response']['format']
            resp_df = self.spark.createDataFrame([], schema = StructType([]))
            if response_type == "parquet":
                resp_df =  self.__extract_parquet()
            return resp_df
        except Exception as exc:
            raise ExtractionException(
                message="Something went wrong while running API Extractor",
                original_exception=exc
            )
        
    def __extract_parquet(self) -> DataFrame:
        response = self.__fetch_response()
        parquet_bytes = BytesIO(response.content)
        tmp_dir = os.path.join(
            self.config.config['DEFAULT']['temp_directory'],
            f'api_parquet_mem_{uuid.uuid4().hex}'
            )
        os.makedirs(tmp_dir, exist_ok=True)
        file_path = f"{tmp_dir}/data.parquet"
        with open(file_path, "wb") as f:
            f.write(parquet_bytes.getbuffer())
        self.logger.info("Parquet loaded in memory at %s", file_path)
        spark_path = file_path.replace("/dbfs", "dbfs:")
        self.logger.info("Reading Parquet into Spark from %s", spark_path)
        return self.spark.read.parquet(spark_path)

    def __fetch_response(self) -> requests.Response:
        cfg = self.config.feed_specs['ingestion_config']
        retry_cfg = cfg.get("retry", {})
        max_attempts = retry_cfg.get("max_attempts", 3)
        backoff_factor = retry_cfg.get("backoff_factor", 2.0)
        max_backoff = retry_cfg.get("max_backoff", 60)
        retry_statuses = set(
            retry_cfg.get("retry_statuses", [429, 500, 502, 503, 504])
        )
        expected_type = cfg.get("response", {}).get("format", "json")
        last_exception = None
        for attempt in range(1, max_attempts + 1):
            try:
                self.logger.info(
                    "API request attempt %s/%s: %s",
                    attempt, max_attempts, cfg['request']["url"]
                )
                resp = cfg['request']
                response = requests.request(
                    method=resp.get("method", "GET"),
                    url=resp["url"],
                    headers=resp.get("headers"),
                    params=resp.get("params"),
                    json=resp.get("body"),
                    timeout=resp.get("timeout", 30),
                    stream=(expected_type == "parquet")
                )
                self.__validate_content_type(response, expected_type)
                if response.status_code < 400:
                    return response
                if response.status_code in retry_statuses:
                    self.__sleep_before_retry(
                        attempt, backoff_factor, max_backoff, response
                    )
                    continue
                response.raise_for_status()
            except RequestException as exc:
                last_exception = exc
                self.logger.warning(
                    "Request error on attempt %s/%s: %s",
                    attempt, max_attempts, exc
                )
                if attempt >= max_attempts:
                    break
                self.__sleep_before_retry(attempt, backoff_factor, max_backoff)

        if last_exception:
            raise last_exception
        raise RuntimeError("API request failed after retries")

    
    def __sleep_before_retry(
        self,
        attempt: int,
        backoff_factor: float,
        max_backoff: int,
        response: requests.Response | None = None
    ) -> None:
        retry_after = None
        if response is not None:
            retry_after = response.headers.get("Retry-After")
        if retry_after:
            sleep_time = min(int(retry_after), max_backoff)
        else:
            sleep_time = min(
                backoff_factor ** attempt + random.uniform(0, 1),
                max_backoff
            )
        self.logger.warning(
            "Retrying in %.2f seconds (attempt %s)",
            sleep_time, attempt
        )
        time.sleep(sleep_time)




from pyspark.sql import SparkSession
import pandas as pd
from handuflow.exception.ValidationError import ValidationError

class ValidationContext:
    spark: SparkSession
    file_hunt_path: str
    raw_json: str
    master_specs_dataframe: pd.DataFrame
    mdf_feed_specs_array: list[dict] | None
    master_spec_name:str

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def get_master_specs(self) -> pd.DataFrame:
        return self.master_specs_dataframe
    
    def _get_table_columns(self, data: dict, rule_name: str) -> set:
        if "source_table_name" not in data:
            raise ValidationError(
                message="Missing 'source_table_name'",
                original_exception=None
            )

        table_name = data["source_table_name"]

        try:
            df = self.spark.table(table_name)
            return {field.name for field in df.schema.fields}
        except Exception as e:
            raise ValidationError(
                message=f"Unable to read schema for table '{table_name}' ",
                original_exception=None
            )




import configparser
from pyspark.sql import SparkSession
from pyspark.sql.functions import expr

from handuflow import Orchestrator

config = configparser.ConfigParser()
config.read("files_dev/config.ini")

spark = (
    SparkSession.builder
    .appName("sdmf")
    .enableHiveSupport()
    .config(
        "spark.jars.packages",
        "io.delta:delta-spark_2.12:3.1.0"
    )
    .config(
        "spark.sql.extensions",
        "io.delta.sql.DeltaSparkSessionExtension"
    )
    .config(
        "spark.sql.catalog.spark_catalog",
        "org.apache.spark.sql.delta.catalog.DeltaCatalog"
    )
    .config("spark.driver.memory", "4g")
    .config("spark.executor.memory", "4g")
    .getOrCreate()
)

# Generate 100 records
df = spark.range(1, 150).toDF("row_id")
df = (
    df
    .withColumn("alpha3_b", expr("concat('USA', row_id)"))
    .withColumn("alpha3_t", expr("concat('US', row_id)"))
    .withColumn("alpha2", expr("substring('US', 1, 2)"))
    .withColumn(
        "english",
        expr("""
            CASE
                WHEN row_id % 4 = 0 THEN 'United States'
                WHEN row_id % 4 = 1 THEN 'Germany'
                WHEN row_id % 4 = 2 THEN 'India'
                ELSE 'Canada'
            END
        """)
    )
    .drop("row_id")
)
print(df.count())
# Recreate table
spark.sql("CREATE DATABASE IF NOT EXISTS demo")
spark.sql("DROP TABLE IF EXISTS demo.customers")
df.write \
    .format("delta") \
    .mode("overwrite") \
    .saveAsTable("demo.customers")


import random
import string

# Function to generate random values
def random_alpha3():
    return ''.join(random.choices(string.ascii_uppercase, k=3))

def random_alpha2():
    return ''.join(random.choices(string.ascii_uppercase, k=2))

def random_english_word():
    return ''.join(random.choices(string.ascii_lowercase, k=6))

# Generate random values
alpha3_b = random_alpha3()
alpha3_t = random_alpha3()
alpha2 = random_alpha2()
english = random_english_word()

# Insert into Spark SQL
query = f"""
INSERT INTO demo.customers (alpha3_b, alpha3_t, alpha2, english)
VALUES ('{alpha3_b}', '{alpha3_t}', '{alpha2}', '{english}')
"""

spark.sql(query)



obj = Orchestrator(config=config, spark=spark)

obj.run()


import os
from handuflow.validation.ValidationRule import ValidationRule
from handuflow.validation.ValidationContext import ValidationContext
from handuflow.exception.ValidationError import ValidationError

class ValidateMasterSpecs(ValidationRule):
    name = "Master spec check"

    def validate(self, context: ValidationContext):
        master_specs = os.path.join(context.file_hunt_path, context.master_spec_name)
        if os.path.exists(master_specs) == False:
            raise ValidationError(
                message = f"System can't find the Master Specs File at [{context.file_hunt_path}]. Terminating process", 
                original_exception=None
            )



import logging
from functools import reduce
import pyspark.sql.functions as F

logger = logging.getLogger(__name__)


class StandardDQExecutor:
    """
    Standalone standard data quality checks.
    All checks return a uniform result structure.
    """

    DEFAULT_PARAMS = {
        "threshold": 0.0,
        "allowed_values": [],
        "range": {"min": float("-inf"), "max": float("inf")},
    }

    def __init__(self, spark):
        self.spark = spark

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------
    def run_check(
        self,
        method_name: str,
        df,
        column=None,
        total_count: int = 0,
        **params,
    ) -> dict:
        """
        Executes a single standard DQ check.
        Returns a uniform result dictionary.
        """

        if not hasattr(self, method_name):
            raise AttributeError(f"Unsupported check method: {method_name}")

        final_params = {**self.DEFAULT_PARAMS, **params}
        self._validate_parameters(final_params)

        method = getattr(self, method_name)

        logger.info(
            "Executing %s with %s on %s",
            method_name,
            self._format_context(method_name, final_params),
            column
        )

        bad_count = method(
            df=df,
            column=column,
            total_count=total_count,
            **final_params,
        )

        threshold = final_params.get("threshold", 0.0)
        passed = self._is_within_threshold(bad_count, total_count, threshold)

        result = {
            "check_name": method_name,
            "column": column,
            "passed": passed,
            "bad_count": bad_count,
            "total_count": total_count,
            "failure_percentage": (
                (float(bad_count) / float(total_count))*float(100) if total_count > 0 else 0.0
            ),
            "failure_ratio": (
                float(bad_count) / float(total_count) if total_count > 0 else 0.0
            ),
            "threshold": threshold,
            "threshold_percentage": float(threshold)*float(100),
        }

        if passed:
            logger.info("%s PASSED on %s", method_name, column)
        else:
            logger.error("%s FAILED on %s", method_name, column)

        return result

    # ------------------------------------------------------------------
    # Parameter Validation
    # ------------------------------------------------------------------
    def _validate_parameters(self, params: dict):
        if not isinstance(params["threshold"], (int, float)):
            raise ValueError("Threshold must be numeric.")
        if not isinstance(params["allowed_values"], list):
            raise ValueError("Allowed values must be a list.")
        if not all(
            isinstance(params["range"][k], (int, float)) for k in ("min", "max")
        ):
            raise ValueError("Range bounds must be numeric.")

    # ------------------------------------------------------------------
    # Logging Helpers
    # ------------------------------------------------------------------
    @staticmethod
    def _format_context(method_name: str, params: dict) -> str:
        if method_name == "_check_value_range":
            r = params["range"]
            return f"range=[{r['min']}, {r['max']}]"
        if method_name == "_check_allowed_values":
            vals = params["allowed_values"]
            preview = vals[:5]
            more = "..." if len(vals) > 5 else ""
            return f"allowed_values={preview}{more}"
        return f"threshold={params.get('threshold')}"

    # ------------------------------------------------------------------
    # Shared Helper
    # ------------------------------------------------------------------
    @staticmethod
    def _is_within_threshold(bad_count, total_count, threshold) -> bool:
        return total_count == 0 or (bad_count / total_count) <= threshold

    # ------------------------------------------------------------------
    # CHECK FUNCTIONS
    # Each check RETURNS bad_count (int)
    # ------------------------------------------------------------------
    def _check_nulls(self, df, column, total_count, **_):
        return df.filter(F.col(column).isNull()).count()

    def _check_value_range(self, df, column, total_count, range, **_):
        return df.filter(
            (F.col(column) < range["min"]) | (F.col(column) > range["max"])
        ).count()

    def _check_allowed_values(self, df, column, total_count, allowed_values, **_):
        return df.filter(~F.col(column).isin(allowed_values)).count()

    def _check_duplicates(self, df, column=None, total_count=0, **_):
        cols = [column] if column else df.columns
        return df.groupBy(cols).count().filter(F.col("count") > 1).count()

    def _check_composite_key(self, df, column, total_count, **_):
        cols = column if isinstance(column, list) else [column]

        nulls = df.filter(
            reduce(lambda x, y: x | y, [F.col(c).isNull() for c in cols])
        ).count()

        dups = df.groupBy(cols).count().filter(F.col("count") > 1).count()

        return nulls + dups

    def _check_primary_key(self, df, column, total_count, **_):
        cols = [column] if isinstance(column, str) else column

        nulls = df.filter(
            reduce(lambda x, y: x | y, [F.col(c).isNull() for c in cols])
        ).count()

        dups = df.groupBy(cols).count().filter(F.col("count") > 1).count()

        return nulls + dups



from handuflow.validation.ValidationRule import ValidationRule
from handuflow.validation.ValidationContext import ValidationContext
from handuflow.exception.ValidationError import ValidationError

class VacuumHoursCheck(ValidationRule):
    name = "Vacuum Hours Check"

    def validate(self, context: ValidationContext):

        if context.mdf_feed_specs_array is None:
            raise ValidationError(
                message="JSON list has not been parsed yet",
                original_exception=None
            )
        
        for json_dict in context.mdf_feed_specs_array:

            if json_dict['data_flow_direction'] != 'SOURCE_TO_BRONZE':

                value = json_dict['feed_specs_dict'].get("vacuum_hours")
                if not isinstance(value, int):
                    raise ValidationError(
                        message=f"'vacuum_hours' must be an integer for feed id {json_dict['feed_id']}",
                        original_exception=None
                    )



# inbuilt
import os
import time
import logging
import configparser
from datetime import datetime
from io import BytesIO
from dataclasses import asdict
import shutil

# external
from openpyxl import Workbook
import pandas as pd

#internal
from handuflow.exception.ResultGenerationException import ResultGenerationException
from handuflow.data_movement_controller.data_class.LoadResult import LoadResult

class ResultGenerator():

    def __init__(
            self, 
            payload: list[dict], 
            file_hunt_path: str, 
            run_id: str, 
            config: configparser.ConfigParser, 
            system_report: pd.DataFrame, 
            load_results: list[LoadResult]
        ) -> None:
        self.payload = payload
        self.file_hunt_path = file_hunt_path
        self.run_id = run_id
        self.final_feed_status = []
        self.standard_results = []
        self.comprehensive_results = []
        self.sheets = []
        self.save_path = None
        self.logger = logging.getLogger(__name__)
        self.config = config

        self.dashboard_cols = [
            "feed_id",
            "feed_name",
            "system_name",
            "subsystem_name",
            "category",
            "sub_category",
            "standard_checks_passed",
            "comprehensive_pre_load_passed",
            "comprehensive_post_load_passed",
            "can_ingest",
        ]

        self.sheets.append(
            {
                "sheet_name": "System Readiness",
                "df":system_report
            }
        )
        self.sheets.append(
            {
                "sheet_name": "Load Report",
                "df": pd.DataFrame([asdict(r) for r in load_results]).sort_values(by="feed_id")
            }
        )

    def run(self):
        try:
            self.logger.info(f"Segregating results...")
            self.__segregate_results()
            self.logger.info(f"Creating sheets...")
            self.__generate_full_feed_status()
            self.logger.info(f"Generating results...")
            self.__generate_standard_results()
            self.logger.info(f"Generate comprehensive results...")
            self.__generate_comprehensive_results()
            self.logger.info(f"Generating dashboard...")
            self.__generate_dashboard()
            self.logger.info(f"Generating final file...")
            self.__generate_result_file()
        except Exception as e:
            raise ResultGenerationException(
                "Something went wrong while running result generator.",
                original_exception=e         
            )

    def __sheet_generator(self, df: pd.DataFrame, sheet_name: str):
        df = df.copy()   # ← THIS is mandatory
        df["run_id"] = self.run_id

        self.sheets.append(
            {
                "sheet_name": sheet_name,
                "df": df
            }
        )

    def __generate_full_feed_status(self):
        final_feed_status_df = pd.DataFrame(self.final_feed_status)
        col = "feed_id"
        final_feed_status_df.insert(0, col, final_feed_status_df.pop(col))
        final_feed_status_df = final_feed_status_df.sort_values(by=col)
        self.__sheet_generator(final_feed_status_df, "Feed Status (B-G)")

    def __generate_standard_results(self):
        df = pd.DataFrame(self.standard_results)
        df = df.explode("standard_checks_result").reset_index(drop=True)
        df = pd.concat(
            [
                df.drop(columns=["standard_checks_result"]),
                df["standard_checks_result"].apply(pd.Series),
            ],
            axis=1,
        )
        df.insert(0, "check_number", range(1, len(df) + 1))
        self.__sheet_generator(df, "Standard Check Result")

    def __generate_comprehensive_results(self):
        if len(self.comprehensive_results) != 0:
            df = pd.DataFrame(self.comprehensive_results)
            df = df.explode("comprehensive_results").reset_index(drop=True)
            df = pd.concat(
                [
                    df.drop(columns=["comprehensive_results"]),
                    df["comprehensive_results"].apply(pd.Series),
                ],
                axis=1,
            )
            df.insert(0, "check_number", range(1, len(df) + 1))
            self.__sheet_generator(df, "Comprehensive Check Result")

    
    def __generate_dashboard(self):
        final_feed_status_df = pd.DataFrame(self.final_feed_status)
        dashboard_df = final_feed_status_df[[c for c in self.dashboard_cols if c in final_feed_status_df.columns]]
        self.__sheet_generator(dashboard_df, "Dashboard")
        

    def __generate_file_name(self) -> str:
        output_directory = os.path.join(self.file_hunt_path, self.config['DEFAULT']['outbound_directory_name'])
        if os.path.exists(output_directory) == False:
            os.mkdir(output_directory)
        now = datetime.now()
        date_str = now.strftime("%Y%m%d")
        return os.path.join(output_directory, f"results_{self.run_id}_{date_str}_{int(time.time())}.xlsx")

    def __normalize_excel_value(self, v):
        if isinstance(v, (list, tuple)):
            return ", ".join(map(str, v))
        return v

    def __generate_result_file(self):
        try:
            wb = Workbook()
            default_sheet = wb.active
            if default_sheet is not None:
                wb.remove(default_sheet)

            for sheet in self.sheets:
                df = sheet["df"]

                ws = wb.create_sheet(title=sheet["sheet_name"][:31])
                ws.append(list(df.columns))
                for row in df.itertuples(index=False, name=None):
                    ws.append([self.__normalize_excel_value(v) for v in row])

            self.save_path = self.__generate_file_name()

            buffer = BytesIO()
            wb.save(buffer)
            buffer.seek(0)

            with open(self.save_path, "wb") as f:
                shutil.copyfileobj(buffer, f)

            self.logger.info(f"Excel report created: {os.path.abspath(self.save_path)}")
        except Exception as e:
            raise ResultGenerationException(
                "Something went wrong while generating excel file.",
                original_exception=e
            )



    def __segregate_results(self):
        try:
            for status in self.payload:
                self.standard_results.append(
                    {
                        "feed_id": status['feed_id'],
                        "standard_checks_result": status['standard_checks_result']
                    }
                )
                self.comprehensive_results.append(
                    {
                        "feed_id": status['feed_id'],
                        "comprehensive_results": status['comprehensive_results']
                    }
                )
                status.pop('standard_checks_result')
                status.pop('comprehensive_results')
                self.final_feed_status.append(status)
                
        except Exception as e:
            raise ResultGenerationException(
                "Something went wrong while segregating results",
                original_exception=e
            )




# inbuilt
import os
import logging
import configparser
from concurrent.futures import ThreadPoolExecutor, as_completed


# external
import pandas as pd
from pyspark.sql import SparkSession

# internal
from handuflow.data_movement_controller.LoadDispatcher import LoadDispatcher
from handuflow.data_movement_controller.data_class.LoadResult import LoadResult


class DataLoadController:

    def __init__(
        self,
        spark: SparkSession,
        allowed_df: pd.DataFrame,
        config: configparser.ConfigParser,
    ) -> None:
        self.logger = logging.getLogger(__name__)
        self.logger.info("Data Load Controller has been initialized...")
        self.master_specs_df = allowed_df
        self.spark = spark
        self.load_results_list = []
        self.config = config

    def run(self):
        self.__prepare()
        self.load_results_list = self.__execute()

    def get_load_results(self) -> list[LoadResult]:
        return self.load_results_list

    def __run_group(self, group) -> list[LoadResult]:
        results = []

        feeds = group["feeds_in_group"]
        group_id = group["parallelism_group_number"]

        self.logger.info(f"Starting group {group_id} with {len(feeds)} feed(s)")

        if len(feeds) > 1:
            with ThreadPoolExecutor(max_workers=len(feeds)) as executor:

                futures = [
                    executor.submit(
                        LoadDispatcher(master_spec=feed, spark=self.spark, config=self.config).dispatch
                    )
                    for feed in feeds
                ]

                for future in as_completed(futures):
                    results.append(future.result())

        else:
            feed = feeds[0]
            results.append(
                LoadDispatcher(master_spec=feed, spark=self.spark, config=self.config).dispatch()
            )

        return results

    def __execute(self) -> list[LoadResult]:
        results = []

        for group in self.execution_groups:
            results.extend(self.__run_group(group))

        return results

    def __prepare(self):
        self.logger.info("Loading validated master specs...")
        self.logger.info("Building ordered execution groups...")

        # Sort first to guarantee execution order
        grouped = self.master_specs_df.sort_values("parallelism_group_number").groupby(
            "parallelism_group_number", sort=False
        )

        self.execution_groups = []

        for key, group in grouped:
            self.logger.info(f"Execution group {key}, Total Feeds: {len(group)}")

            self.execution_groups.append(
                {
                    "parallelism_group_number": key,
                    "feeds_in_group": group.to_dict(orient="records"),
                }
            )



from handuflow.exception.BaseException import BaseException

class ResultGenerationException(BaseException):
    def __init__(self, message=None, details=None, original_exception=None):
        super().__init__(
            message or "Result Generation Exception",
            details=details,
            original_exception=original_exception
        )


# inbuilt
import os
import logging
import configparser
from collections import defaultdict

# external
import matplotlib

import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle, FancyArrow
import pandas as pd

class DataFlowDiagramGenerator():

    def __init__(self, validated_dataframe: pd.DataFrame, config: configparser.ConfigParser, run_id: str) -> None:
        self.config = config
        self.run_id = run_id
        self.logger = logging.getLogger(__name__)
        self.BOX_WIDTH = float(config['LINEAGE_DIAGRAM']['BOX_WIDTH'])
        self.BOX_HEIGHT = float(config['LINEAGE_DIAGRAM']['BOX_HEIGHT'])
        self.X_GAP = float(config['LINEAGE_DIAGRAM']['X_GAP'])
        self.Y_GAP = float(config['LINEAGE_DIAGRAM']['Y_GAP'])
        self.ROOT_GAP = float(config['LINEAGE_DIAGRAM']['ROOT_GAP'])
        self.validated_dataframe = validated_dataframe
        self.name_map = {}
        self.current_y = 0
        self.positions = {}
        self.lineage_bounds = []
        self.roots = None
        self.lineage_id = 1
        matplotlib.use("Agg")


    def run(self):
        if len(self.validated_dataframe) != 0:
            self.logger.info('Building tree structure...')
            self.__build_tree_structure()
            self.logger.info('Layout each root separately...')
            self.__layout_root()
            self.logger.info('Plot...')
            self.__plot()
            self.logger.info('Draw lineage separators & labels...')
            self.__lineage_separators_and_lables()
            self.logger.info('Final cleanup and generate files...')
            self.__title_cleanup_and_save()

    def __build_tree_structure(self):
        self.children = defaultdict(list)
        self.name_map = dict(zip(self.validated_dataframe['feed_id'], self.validated_dataframe['feed_name']))
        for _, r in self.validated_dataframe.iterrows():
            self.children[r.parent_feed_id].append(r.feed_id)

    def __layout_subtree(self, node_id, depth):
        for child in self.children.get(node_id, []):
            self.__layout_subtree(child, depth + 1)
        if node_id not in self.positions:
            self.positions[node_id] = (
                depth * (self.BOX_WIDTH + self.X_GAP),
                self.current_y
            )
            self.current_y -= self.Y_GAP
        child_ys = [
            self.positions[c][1]
            for c in self.children.get(node_id, [])
            if c in self.positions
        ]
        if child_ys:
            avg_y = sum(child_ys) / len(child_ys)
            self.positions[node_id] = (
                depth * (self.BOX_WIDTH + self.X_GAP),
                avg_y
            )

    def __layout_root(self):
        self.roots = self.children[0]
        for root in self.roots:
            lineage_top = self.current_y
            self.__layout_subtree(root, 0)
            lineage_bottom = self.current_y
            self.lineage_bounds.append((lineage_top, lineage_bottom, self.lineage_id))
            self.lineage_id += 1
            self.current_y -= self.ROOT_GAP

    def __plot(self):
        fig, self.ax = plt.subplots(figsize=(80, max(10, len(self.validated_dataframe) * 2)))
        for fid, (x, y) in self.positions.items():
            if fid in self.roots:
                facecolor = "#C0392B"   # red
                text_color = "white"
            else:
                facecolor = "#D6EAF8"   # default blue
                text_color = "black"

            rect = Rectangle(
                (x, y),
                self.BOX_WIDTH,
                self.BOX_HEIGHT,
                edgecolor="black",
                facecolor=facecolor,
                linewidth=1.6
            )
            self.ax.add_patch(rect)

            label = f"{fid}: {self.name_map[fid]}"

            self.ax.text(
                x + self.BOX_WIDTH / 2,
                y + self.BOX_HEIGHT / 2,
                label,
                ha="center",
                va="center",
                fontsize=20,
                fontweight="bold",
                color=text_color,
                wrap=True
            )

        for _, r in self.validated_dataframe.iterrows():
            if r.parent_feed_id != 0:
                px, py = self.positions[r.parent_feed_id]
                cx, cy = self.positions[r.feed_id]

                arrow = FancyArrow(
                    px + self.BOX_WIDTH,
                    py + self.BOX_HEIGHT / 2,
                    cx - (px + self.BOX_WIDTH),
                    cy - py,
                    width=0.03,
                    head_width=0.4,
                    head_length=0.5,
                    length_includes_head=True,
                    color="black"
                )
                self.ax.add_patch(arrow)

    def __lineage_separators_and_lables(self):
        x_min = -2
        x_max = max(x for x, _ in self.positions.values()) + self.BOX_WIDTH + 2
        for top_y, bottom_y, lid in self.lineage_bounds:
            y_sep = bottom_y + self.ROOT_GAP / 2
            self.ax.hlines(
                y=y_sep,
                xmin=x_min,
                xmax=x_max,
                colors="gray",
                linestyles="dashed",
                linewidth=2
            )
            self.ax.text(
                x_min + 0.5,
                y_sep + 0.3,
                f"Lineage {lid}",
                fontsize=16,
                fontweight="bold",
                color="gray"
            )

    def __title_cleanup_and_save(self):
        self.ax.set_title(
            "handuflow - Feed Flow Diagram",
            fontsize=28,
            fontweight="bold",
            pad=30
        )
        self.ax.axis("off")
        self.ax.autoscale_view()
        path = os.path.join(self.config['DEFAULT']['file_hunt_path'], self.config['DEFAULT']['outbound_directory_name'], f'feed_lineage_{self.run_id}.png')
        plt.savefig(path)
        plt.close()


# inbuilt
import logging

# external
from pyspark.sql import SparkSession
from pyspark.sql.types import StructType

# internal
from handuflow.data_movement_controller.data_class.LoadResult import LoadResult
from handuflow.data_movement_controller.data_class.LoadConfig import LoadConfig
from handuflow.data_movement_controller.BaseLoadStrategy import BaseLoadStrategy
from handuflow.exception.DataLoadException import DataLoadException

class FullLoad(BaseLoadStrategy):

    def __init__(self, config: LoadConfig, spark: SparkSession) -> None:
        super().__init__(config=config, spark=spark)
        self.logger = logging.getLogger(__name__)
        self.logger.info("Initializing FULL_LOAD data transfer component...")

    def load(self) -> LoadResult:
        """
        Perform a FULL overwrite load into the target table in Databricks Unity Catalog.
        This replaces all existing data in the target with data from the staging layer.

        Returns:
            LoadResult: LoadResult.
        """
        try:
            self._enforce_load_type_consistency()
            is_staging_layer_created = self._create_staging_layer()
            
            if not is_staging_layer_created:
                self.logger.error(
                    f"FULL LOAD failed for {self._current_target_table_name}: Staging Layer wasn't created."
                )
                return LoadResult(
                    feed_id = self.config.master_specs['feed_id'], 
                    success=False,
                    total_rows_inserted=0, 
                    total_rows_deleted=0, 
                    total_rows_updated=0
                )
            full_staging_df = self._current_staging_table_df
            record_count = full_staging_df.count()
            if full_staging_df is None or record_count == 0:
                self.logger.info(
                    f"FULL LOAD skipped: staging DataFrame is empty for {self._current_target_table_name}."
                )
                return LoadResult(
                    feed_id = self.config.master_specs['feed_id'], 
                    success=True, 
                    skipped=True, 
                    total_rows_inserted=0, 
                    total_rows_deleted=0,
                    total_rows_updated=0
                )
            if self.spark.catalog.tableExists(self._current_target_table_name) == True:
                previous_count = self.spark.table(self._current_target_table_name).count()
            else:
                previous_count = 0
            full_staging_df = self._enforce_schema(full_staging_df, StructType.fromJson(self.config.feed_specs['selection_schema']))
            (
                full_staging_df.write.
                format("delta")
                .mode("overwrite")
                .option("overwriteSchema", "true")
                .saveAsTable(self._current_target_table_name)
            )
            self.logger.info(
                f"FULL LOAD completed successfully for {self._current_target_table_name} "
                f"({record_count} records loaded)."
            )
            return LoadResult(
                feed_id = self.config.master_specs['feed_id'],
                success=True, 
                total_rows_inserted=record_count,
                total_rows_updated=0,
                total_rows_deleted=previous_count
            )
        except Exception as e:
            raise DataLoadException(
                message=f"Feed ID: {self.config.master_specs['feed_id']}, Error during FULL LOAD for {self._current_target_table_name}: {str(e)}",
                original_exception=e
            )



# inbuilt
import os
import logging
import configparser
from datetime import datetime, timedelta


# external
import pandas as pd


class SystemCleanup:
    def __init__(self, config: configparser.ConfigParser, master_specs: pd.DataFrame) -> None:
        self.logger = logging.getLogger(__name__)
        self.logger.info("Starting system cleanup process...")
        self.config = config
        self.master_specs = master_specs
        self.retention_days = int(config["DEFAULT"]["retention_policy_in_days"])
        self.final_log_dir = os.path.join(
            config["DEFAULT"]["log_base_path"],
            config["DEFAULT"]["log_directory_name"],
        )
        temp = config["DEFAULT"]["temp_directory"]
        temp = temp.replace("/dbfs", "")
        self.temp_log_dir = os.path.join(
            temp,
            config["DEFAULT"]["log_directory_name"],
        )

        self.outbound_dir = os.path.join(
            config["DEFAULT"]["file_hunt_path"],
            config["DEFAULT"]["outbound_directory_name"],
        )

    def run(self):
        self.logger.info("Removing old logs...")
        self.__remove_old_logs()
        self.logger.info("Removing old outputs...")
        self.__remove_old_outputs()

    def __remove_old_logs(self):
        cutoff = datetime.now() - timedelta(days=self.retention_days)
        for f in os.listdir(self.final_log_dir):
            path = os.path.join(self.final_log_dir, f)
            if os.path.isfile(path):
                if datetime.fromtimestamp(os.path.getmtime(path)) < cutoff:
                    os.remove(path)
                    self.logger.warning(f"Removed old log file: {path}")
        for f in os.listdir(self.temp_log_dir):
            path = os.path.join(self.temp_log_dir, f)
            if os.path.isfile(path):
                os.remove(path)
                self.logger.warning(f"Removed old log file: {path}")

    def __remove_old_outputs(self):
        cutoff = datetime.now() - timedelta(days=self.retention_days)
        for f in os.listdir(self.outbound_dir):
            path = os.path.join(self.outbound_dir, f)
            if os.path.isfile(path):
                if datetime.fromtimestamp(os.path.getmtime(path)) < cutoff:
                    os.remove(path)
                    self.logger.warning(f"Removed old output file: {path}")

    def __enforce_delta_retention_for_all_tables(self):
        pass



from handuflow.validation.ValidationRule import ValidationRule
from handuflow.validation.ValidationContext import ValidationContext
from handuflow.exception.ValidationError import ValidationError

class PartitionKeysCheck(ValidationRule):
    name = "Partition keys check"

    def validate(self, context: ValidationContext):

        if context.mdf_feed_specs_array is None:
            raise ValidationError(
                message="JSON list has not been parsed yet",
                original_exception=None
            )
        
        for json_dict in context.mdf_feed_specs_array:

            if json_dict['data_flow_direction'] != 'SOURCE_TO_BRONZE':
                
                data = json_dict['feed_specs_dict']
                value = data.get("partition_keys")
                if "partition_keys" not in data:
                    raise ValidationError(
                        message=f"Missing 'partition_keys' for feed id {json_dict['feed_id']}",
                        original_exception=None
                    )
                if not isinstance(value, list):
                    raise ValidationError(
                        message=f"'partition_keys' must be an list for feed id {json_dict['feed_id']}",
                        original_exception=None
                    )

                table_name = data["source_table_name"]
                if context.spark.catalog.tableExists(table_name):
                    table_columns = context._get_table_columns(data, self.name)
                    for key in value:
                        if not isinstance(key, str):
                            raise ValidationError(
                                message=f"'partition_keys' must contain only strings for feed id {json_dict['feed_id']}",
                                original_exception=None
                            )
                        if key not in table_columns:
                            raise ValidationError(
                                message=f"'partition_keys' column '{key}' not found in table "
                                        f"'{data['source_table_name']}'  for feed id {json_dict['feed_id']}",
                                original_exception=None
                            )



# inbuilt
import logging
import configparser

# external
from pyspark.sql import SparkSession
import pandas as pd

# internal
from handuflow.validation.Validator import Validator
from handuflow.validation.ValidationContext import ValidationContext
from handuflow.validation.validation_rules.ValidateMasterSpecs import ValidateMasterSpecs
from handuflow.validation.validation_rules.EnforceMasterSpecsStructure import (
    EnforceMasterSpecsStructure,
)
from handuflow.validation.validation_rules.ValidateFeedSpecsJSON import ValidateFeedSpecsJSON
from handuflow.validation.validation_rules.PrimaryKey import PrimaryKey
from handuflow.validation.validation_rules.ColumnExistsInSelection import (
    ColumnExistsInSelection,
)
from handuflow.validation.validation_rules.EnforceStandardChecks import EnforceStandardChecks
from handuflow.validation.validation_rules.StandardCheckStructureCheck import (
    StandardCheckStructureCheck,
)
from handuflow.validation.validation_rules.ComprehensiveChecksDependencyDatasetCheck import (
    ComprehensiveChecksDependencyDatasetCheck,
)
from handuflow.validation.validation_rules.PartitionKeysCheck import PartitionKeysCheck
from handuflow.validation.validation_rules.CompositeKeysCheck import CompositeKeysCheck
from handuflow.validation.validation_rules.VacuumHoursCheck import VacuumHoursCheck
from handuflow.exception.ValidationError import ValidationError


class SystemLaunchValidator:

    def __init__(
        self,
        file_hunt_path: str,
        spark: SparkSession,
        config: configparser.ConfigParser,
    ) -> None:
        self.logger = logging.getLogger(__name__)
        self.spark = spark
        self.config = config

        self.context = ValidationContext(
            spark=spark,
            file_hunt_path=file_hunt_path,
            master_spec_name=self.config["FILES"]["master_spec_name"],
        )

    def __init_rules(self):
        self.rules = [
            ValidateMasterSpecs(),
            EnforceMasterSpecsStructure(),
            ValidateFeedSpecsJSON(),
            PrimaryKey(),
            ColumnExistsInSelection(),
            PartitionKeysCheck(),
            CompositeKeysCheck(),
            VacuumHoursCheck(),
            EnforceStandardChecks(),
            StandardCheckStructureCheck(),
            ComprehensiveChecksDependencyDatasetCheck(),
        ]

    def run(self):
        try:
            self.__init_rules()
            validator = Validator(self.rules, fail_fast=True)
            return validator.validate(self.context)
        except Exception as e:
            raise ValidationError(
                    "Something went wrong in system validation",
                    original_exception=e
                )

    def get_validated_master_specs(self) -> pd.DataFrame:
        opt_df = self.context.get_master_specs()
        opt_df = opt_df[opt_df["is_active"] == True]
        return opt_df



# inbuilt
import uuid
import logging
import configparser

# external
import pandas as pd
from pyspark.sql import SparkSession

# internal
from handuflow.config.LoggingConfig import LoggingConfig
from handuflow.exception.SystemError import SystemError
from handuflow.result_generator.ResultGenerator import ResultGenerator
from handuflow.validation.SystemLaunchValidator import SystemLaunchValidator
from handuflow.data_quality.runner.FeedDataQualityRunner import FeedDataQualityRunner
from handuflow.data_movement_controller.DataLoadController import DataLoadController
from handuflow.system_cleanup.SystemCleanup import SystemCleanup
from handuflow.data_flow_diagram_generator.DataFlowDiagramGenerator import (
    DataFlowDiagramGenerator,
)


class Orchestrator:

    def __init__(self, spark: SparkSession, config: configparser.ConfigParser) -> None:
        self.config = config
        self.run_id = uuid.uuid4().hex
        self.my_LoggingConfig = LoggingConfig(run_id=self.run_id, config=config)
        self.my_LoggingConfig.configure()
        self.logger = logging.getLogger(__name__)
        self.logger.info(
            f"""handuflow - Reliable data movement and evolution at scale"""
        )
        self.spark = spark
        self.file_hunt_path = config["DEFAULT"]["file_hunt_path"]
        self.system_run_report = pd.DataFrame()
        self.logger.info(f"Current Run Id: {self.run_id}")
        # self.logger.info(
        #     f'spark.scheduler.mode: {spark.sparkContext.getConf().get("spark.scheduler.mode")}'
        # )

    def __system_prerequisites(self) -> bool:
        my_SystemLaunchValidator = SystemLaunchValidator(
            file_hunt_path=self.file_hunt_path, spark=self.spark, config=self.config
        )
        validation_result = my_SystemLaunchValidator.run()
        self.validated_master_specs_df = (
            my_SystemLaunchValidator.get_validated_master_specs()
        )
        self.validated_master_specs_df2 = (
            my_SystemLaunchValidator.get_validated_master_specs()
        )
        self.system_run_report = validation_result.results_df
        return validation_result.passed

    def run(self):
        self.logger.info("Ensuring system readiness...")
        self.logger.warning(
            "All feeds from external sources will be dumped to the bronze (raw) layer, their validation may be skipped."
        )
        is_system_ready = self.__system_prerequisites()
        if is_system_ready == True:

            self.logger.info("System is up and ready.")
            self.logger.info("Validating and loading data...")
            self.__validate_and_load()
            self.logger.info("Generating lineage diagram...")
            self.__generate_lineage_diagram()
            self.logger.info("System has finished processing this batch.")
            my_SystemCleanup = SystemCleanup(config=self.config, master_specs=self.validated_master_specs_df2)
            my_SystemCleanup.run()
            self.logger.warning(
                "Saving logs to specified final log directory, no logs after this point will be retained in *.log file."
            )
            self.logger.info("==FINAL LOG==")
            self.my_LoggingConfig.move_logs_to_final_location()
            self.logger.info("System has finished processing data.")
            self.logger.info("Thanks for using SDMF.")
        else:
            self.logger.error("System is not ready, feeds will not be processed.")

    def __generate_lineage_diagram(self):
        try:
            my_DataFlowDiagramGenerator = DataFlowDiagramGenerator(
                validated_dataframe=self.validated_master_specs_df2,
                config=self.config,
                run_id=self.run_id,
            )
            my_DataFlowDiagramGenerator.run()
        except Exception as e:
            raise SystemError(
                "Something went wrong while generating Lineage diagram",
                details={},
                original_exception=e,
            )

    def __validate_and_load(self):
        if self.validated_master_specs_df is None:
            raise SystemError(
                message="SDMF was not able to find validated specs.",
                details=None,
                original_exception=None,
            )
        extraction_df = self.validated_master_specs_df[
            self.validated_master_specs_df["data_flow_direction"] == "SOURCE_TO_BRONZE"
        ]
        load_results = []
        if len(extraction_df) != 0:
            self.logger.info("Performing extraction as configured...")
            my_DataLoadController = DataLoadController(
                allowed_df=extraction_df, spark=self.spark, config=self.config
            )
            my_DataLoadController.run()
            res = my_DataLoadController.get_load_results()
            load_results.extend(res)
        self.validated_master_specs_df = self.validated_master_specs_df[
            self.validated_master_specs_df["data_flow_direction"] != "SOURCE_TO_BRONZE"
        ]
        obj = FeedDataQualityRunner(
            self.spark, self.validated_master_specs_df.to_dict(orient="records")
        )
        obj.run()
        pre_load_mainifest = obj._finalize()
        can_ingest_feed_id = []
        for ready_for_ingest in pre_load_mainifest:
            if ready_for_ingest["can_ingest"] == True:
                can_ingest_feed_id.append(ready_for_ingest["feed_id"])
        self.logger.info(f"Valid Feed ID's: {can_ingest_feed_id}")
        if len(can_ingest_feed_id) > 0:
            allowed_df = self.validated_master_specs_df[
                self.validated_master_specs_df["feed_id"].isin(can_ingest_feed_id)
            ]
            my_DataLoadController = DataLoadController(
                allowed_df=allowed_df, spark=self.spark, config=self.config
            )
            my_DataLoadController.run()
            res2 = my_DataLoadController.get_load_results()
            load_results.extend(res2)
            obj.adhoc_post_load()
            all_feed_manifest = obj._finalize()
            my_ResultGenerator = ResultGenerator(
                all_feed_manifest,
                file_hunt_path=self.file_hunt_path,
                run_id=self.run_id,
                config=self.config,
                system_report=self.system_run_report,
                load_results=load_results,
            )
            my_ResultGenerator.run()
        else:
            self.logger.warning(
                f"No non-Extraction feeds passed their defined validation. Please check the run report in SDMF outbound directory with run id: [{self.run_id}]"
            )



from dataclasses import dataclass


@dataclass
class FeedDQSummaryRow:
    feed_name: str
    feed_ok: bool
    standard_checks_passed: bool
    pre_load_error: bool
    pre_load_warning_count: int
    executed_pre_load_checks: int